#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" kberylsettings.beryl -> wrappers around berylsettings objects.

"""
import berylsettings

from kdecore import KIcon, locate
from kberylsettings.lib import Signals, icon, iconLoader
from qt import QImage, QPixmap, QObject


class Context(QObject):
    """ Context -> wraps berylsetting.Context instances with extra
    methods and properties.

    This class inherits QObject so that instances can emit Qt signals.
    """
    activePluginsSettingName = 'active_plugins'

    def __init__(self, context=None):
        QObject.__init__(self)
        if context is None:
            context = berylsettings.Context()
            context.Read()
        self.context = context

    def __getattr__(self, value):
        """ x.__getattr__('name') <==> x.name

        """
        return getattr(self.context, value)

    def __repr__(self):
        """ x.__repr__() <==> repr(x)

        """
        args = (self.__module__, self.__class__.__name__, hex(abs(id(self))))
        return "<%s.%s instance at %s>" % args
    __str__ = __repr__
    
    def getCategories(self):
        """ sorted categories sequence

        @return sorted list of Category instances
        """
        seq = [Category(c) for c in self.context.Categories]
        seq.sort()
        return seq
    categories = property(getCategories)

    def getPlugins(self):
        """ sorted plugins sequence

        @return sequence of Plugin wrapper instances
        """
        seq = [Plugin(p) for p in self.context.Plugins]
        seq.sort()
        return seq
    plugins = property(getPlugins)

    def plugin(self, value):
        """ locates plugin by name, description, or index

        @param value plugin name, short description, or index
        @return Plugin wrapper instance
        """
        try:
            value + 0
        except (TypeError, ):
            for plugin in self.plugins:
                if value == plugin.ShortDesc:
                    return Plugin(plugin)
                if value == plugin.Name:
                    return Plugin(plugin)
        else:
            return list(self.plugins)[value]

    def getActive(self):
        """ sequence of active plugin names

        @return sequence of active plugin names
        """
        act = self.general.Setting(self.activePluginsSettingName).Value
        return act

    def setActive(self, active):
        """ sets active plugins

        @param active sequence of plugin names to set as active
        @return None
        """
        self.general.Setting(self.activePluginsSettingName).Value = active
    active = property(getActive, setActive)

    def getGeneral(self):
        """ general plugin instance

        @return Plugin wrapper instance
        """
        return Plugin(self.context.Plugin(Plugin.generalName))
    general = property(getGeneral)

    def write(self):
        """ saves the beryl context

        @return None
        """
        self.emit(Signals.statusMessage, ('Saving Beryl settings....', ))
        self.context.Write()
        self.reload()

    def reload(self):
        """ messages the extension to reload the new settings

        @return None
        """
        berylsettings.send_reload()
        self.emit(Signals.statusMessage, ('Beryl settings reloaded.', ))    


class Category:
    """ Category -> wraps berylsetting.Category instances with extra
    methods and properties.

    """
    iconNameMap = {
        'wm' : 'kwin',
        'effects' : 'style',
        'misc' : 'package_toys',
        'settings' : 'configure',
        'imageformat' : 'icons',
        'devel' : 'source',
        'desktop' : 'desktop',
        'accessibility' : 'access',
    }
    
    def __init__(self, category):
        self.category = category

    def __cmp__(self, other):
        """ x.__cmp__(y) <==> cmp(x, y)

        """
        return cmp(self.ShortDesc, other.ShortDesc)

    def __getattr__(self, value):
        """ x.__getattr__('name') <==> x.name

        """
        return getattr(self.category, value)

    def icon(self, size):
        """ icon for this category

        @param size KIcon size
        @return icon instance
        """
        name = self.iconNameMap.get(self.Name, self.Name)
        return icon(name, KIcon.NoGroup, size)

    def plugins(self):
        """ sequence of plugin instances in this category

        @return sequence of active plugin names
        """
        seq = [Plugin(p) for p in self.category.Plugins]
        seq.sort()
        return seq
    plugins = property(plugins)


class Plugin:
    """ Plugin -> wraps berylsetting.Plugin instances with extra
    methods and properties.

    """
    generalName = '_'
    iconCache = {}
    
    def __init__(self, plugin):
        self.plugin = plugin

    def __cmp__(self, other):
        """ x.__cmp__(y) <==> cmp(x, y)

        """
        if other is None:
            return -1
        if not isinstance(other, Plugin):
            other = Plugin(other)
        if self.isGeneral:
            return -1
        if other.isGeneral:
            return 1
        return cmp(self.plugin.ShortDesc, other.plugin.ShortDesc)

    def __getattr__(self, value):
        """ x.__getattr__('name') <==> x.name

        """
        return getattr(self.plugin, value)

    def __repr__(self):
        """ x.__repr__() <==> repr(x)

        """
        args = (self.__module__, self.__class__.__name__,
                self.Name,
                hex(abs(id(self))))
        return "<%s.%s instance name='%s' at %s>" % args
    __str__ = __repr__
    
    def isGeneral(self):
        """ true if this plugin is a general plugin

        """
        return self.plugin.Name == self.generalName
    isGeneral = property(isGeneral)

    def icon(self, size, loader):
        """ icon for this plugin

        @param size KIcon size
        @param loader KIconLoader instance
        @return icon instance
        """
        name = self.plugin.Name
        try:
            pix = self.iconCache[(name, size)]
        except (KeyError, ):
            path = locate('icon', 'beryl-settings-section-%s.svg' % name)
            if not path:
                path = 'unknown'
            ico = loader.loadIcon(path, KIcon.NoGroup)
            img = ico.convertToImage()
            pix = self.iconCache[(name, size)] = QPixmap()
            pix.convertFromImage(img.smoothScale(size, size, QImage.ScaleMin))
        return pix

    def iconPath(self):
        """ path to icon for this plugin

        @return icon path
        """
        name = self.plugin.Name        
        return locate('icon', 'beryl-settings-section-%s.png' % name)
            
    altType = {
        'Int' : 'Values',
        'Float' : 'Values',
        'Color' : 'Colors',
        'Bool' : 'Choices',
        'String' : 'Choices',
        'Binding' : 'Bindings',
        'List of String' : 'Items',
        }

    altLong = {
        'Values' : 'Configure numeric values',
        'Colors' : 'Choose colors',
        'Choices' : 'Turn things on or off',
        'Bindings' : 'Connect keyboard, mouse, edges and the system bell',
        'Items' : 'Select items from (and into) lists of strings',
        }

    def settingTypeDesc(self, key):
        """ lookup a long description for alternate type name

        @param key alternate type name (keys of altType dict)
        @return string description or key
        """
        return self.altLong.get(key, key)
    
    def settings(self):
        """ mapping of plugin settings

        """
        settingtype = {}
        for setting in [Setting(s) for s in self.plugin.Settings]:
            typ = setting.Type
            key = self.altType.get(typ, typ)
            settingtype.setdefault(key, []).append(setting)
        return settingtype
    settings = property(settings)

    def groups(self):
        """ sorted groups sequence

        """
        groups = [Group(g) for g in self.plugin.Groups]
        groups.sort()
        return groups
    groups = property(groups)


class Group:
    """ Group -> wraps berylsetting.Group instances with extra
    methods and properties.

    """
    def __init__(self, group):
        self.group = group

    def __getattr__(self, value):
        """ x.__getattr__('name') <==> x.name

        """
        return getattr(self.group, value)

    def __cmp__(self, other):
        """ x.__cmp__(y) <==> cmp(x, y)

        """
        if not isinstance(other, Group):
            other = Group(other)
        return cmp(self.label, other.label)

    def label(self):
        """ label for this group

        """
        return self.Name or self.Desc or 'Uncategorized'
    label = property(label)

    def settings(self, plugin):
        """ sequence of settings in this group for plugin

        """
        group = self.Name
        return [Setting(s) for s in plugin.Settings if s.Group==group]


class Setting:
    """ Setting -> wraps berylsetting.Setting instances with extra
    methods and properties.

    """
    iconNameMap = {
        'Bindings' : 'mouse',
        'Choices': 'apply',        
        'Colors' : 'colorize',
        'Items': 'edit',
        'Values' : 'exec',
        'Back' : 'back',
        'None' : 'back',
    }

    def __init__(self, setting):
        self.setting = setting

    def __getattr__(self, value):
        """ x.__getattr__('name') <==> x.name

        """
        return getattr(self.setting, value)

    def __repr__(self):
        """ x.__repr__() <==> repr(x)

        """
        args = (self.__module__, self.__class__.__name__,
                self.Name,
                hex(abs(id(self))))
        return "<%s.%s instance name='%s' at %s>" % args
    __str__ = __repr__
    
    def set(self, value):
        """ set the setting's value

        @param value new value for setting
        @return None
        """
        self.setting.Value = value
        
    def icon(cls, value, size):
        """ icon for setting type

        @param value setting type name
        @param size KIcon size
        @return icon instance
        """
        name = cls.iconNameMap.get(str(value), cls.iconNameMap['Choices'])
        return icon(name, size=size)
    icon = classmethod(icon)

    def iconPath(cls, value):
        """ path to icon for setting type

        @param value setting type name
        @return icon path
        """
        name = cls.iconNameMap.get(str(value), cls.iconNameMap['Choices'])
        ico = iconLoader().iconPath(name, KIcon.NoGroup)
        return ico
    iconPath = classmethod(iconPath)
    
    def getValue(self):
        """ get value of setting, explicitly silencing errors

        @return setting value
        """
        try:
            return self.setting.Value
        except (IndexError, ):
            return ''

    def setValue(self, value):
        """ set value of setting

        @param new value for setting
        @return None
        """
        self.set(value)
    Value = property(getValue, setValue)
