#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" kberylsettings.contentframe -> defines the right side about/setting views.

"""
from kdecore import KIcon, i18n
from kdeui import KDialog, KMessageBox, KPassivePopup, \
     KSeparator, KStdGuiItem, KTabWidget
from qt import Qt, QFrame, QHBoxLayout, QLabel, QScrollView, QSize, \
     QSizePolicy, QStringList, QToolTip, QVBoxLayout

from kberylsettings.aboutpage import AboutPage
from kberylsettings.beryl import Setting
from kberylsettings.lib import Signals, icon, iconLoader, iconSet
from kberylsettings.widget import Frame, SmallPushButton, WidgetStack, guiButton
from kberylsettings.settingwidget import settingWidget


marginHint = KDialog.marginHint()
spacingHint = KDialog.spacingHint()


class ContentFrame(WidgetStack):
    """ ContentFrame -> multiple displays of plugin information

    This widget stack builds and manages three types of pages:

        1.  an 'about' page for displaying plugin information.

        2.  a 'single' page for displaying a setting or a
        sequence of settings.
      
        3.  a 'multiple' page for displaying settings in groups
        via a tabbed interface.
        
    """
    aboutPageId, singlePageId, tabPageId = range(3)

    def __init__(self, parent):
        WidgetStack.__init__(self, parent)
        self.aboutPage = aboutPage = AboutPage(self)
        self.singlePage = singlePage = SinglePage(self)
        self.tabPage = tabPage = TabPage(self)

        addWidget = self.addWidget
        addWidget(aboutPage, self.aboutPageId)
        addWidget(singlePage, self.singlePageId)
        addWidget(tabPage, self.tabPageId)
        
        root = self.topLevelWidget()
        connect = self.connect
        connect(root, Signals.showAbout, self.showAboutPage)        
        connect(root, Signals.showSettings, self.showSinglePage)
        connect(root, Signals.showGroups, self.showTabPage)
        connect(aboutPage, Signals.selectPrevious,
                root, Signals.selectPrevious)
        
    def showAboutPage(self, context, plugin):
        """ displays the about page

        @param plugin berylsettings Plugin instance
        @return None
        """
        self.aboutPage.showAbout(context, plugin)
        self.raiseWidget(self.aboutPageId)

    def showSinglePage(self, plugin, arg):
        """ displays the page for a single type of setting

        @param plugin berylsettings Plugin instance
        @return None
        """
        self.singlePage.showSettings(plugin, arg)        
        self.raiseWidget(self.singlePageId)

    def showTabPage(self, plugin):
        """ displays the page for groups of settings

        @param plugin berylsettings Plugin instance
        @return None
        """
        self.tabPage.setGroups(plugin)
        self.raiseWidget(self.tabPageId)

    def queryClose(self):
        """ checks for unsaved changes in a setting page

        @return True if no changes, changes discarded, or changes
        applied; False if canceled
        """
        try:
            w = self.visibleWidget()
            return w.unsavedCheck(None, None)
        except (AttributeError, ), exc:
            return True


class SettingPage:
    """ SettingPage -> mixin with methods common to both the
    SinglePage and the TabPage types.

    """
    previous = (None, None)
    unsavedText = ('There are unsaved changes in the active settings.\n'
                   'Do you want to apply the changes before changing views '
                   'or discard the changes?')

    def buildConnections(self):
        """ build the connections for this instance

        @return None
        """
        connect = self.connect
        connect(self.footer.helpButton, Signals.clicked, self.settingHelp)
        connect(self.footer.defaultsButton, Signals.clicked, self.settingDefaults)
        connect(self.footer.resetButton, Signals.clicked, self.settingsReset)
        connect(self.footer.applyButton, Signals.clicked, self.settingApply)        
        root = self.topLevelWidget()
        connect(self, Signals.berylSettingChanged, root.onContextChanged)
        connect(self, Signals.statusMessage, root.showMessage)
        connect(self, Signals.selectPrevious, root, Signals.selectPrevious)

    def onSomethingChanged(self):
        """ enable apply and reset buttons on setting value change

        @return None
        """
        self.enableApplyReset(True, True)

    def enableApplyReset(self, enableApply, enableReset):
        """ convenience for setting enabled state of apply and reset buttons

        @return None
        """
        self.footer.applyButton.setEnabled(enableApply)
        self.footer.resetButton.setEnabled(enableReset)

    def unsaved(self):
        """ checks for unsaved items (defers to apply button enabled state)

        @return True if unsaved settings
        """
        return self.footer.applyButton.isEnabled()

    def unsavedDialog(self):
        """ display a dialog warning of unsaved changes

        @return KMessageBox.Yes, .No, or .Cancel
        """
        msg = KMessageBox.warningYesNoCancel
        return msg(self, self.unsavedText,
                   i18n('Unsaved Changes'),
                   KStdGuiItem.apply(),
                   KStdGuiItem.discard())

    def unsavedCheck(self, plugin, arg):
        """ checks for unsaved items and prompts for action if any

        @param plugin berylsettings Plugin instance
        @param arg setting name or setting section name
        @return True if no save needed, changes discarded or applied;
        False otherwise
        """
        ret = True
        if self.unsaved():
            res = self.unsavedDialog()
            if res == KMessageBox.Cancel:
                self.emit(Signals.selectPrevious, self.previous)
                ret = False
            elif res == KMessageBox.Yes:
                self.settingApply()
            else:
                self.settingDiscard()
        if ret:
            self.previous = (plugin, arg)
        return ret

    def settingHelp(self):
        """ not implemented

        """

    def settingDiscard(self):
        """ unsaved changes discarded; disable apply and reset buttons

        @return None
        """
        self.enableApplyReset(False, False)

    def settingDefaults(self):
        """ set each setting to its default value

        @return None
        """
        excs = []
        for w in self.queryList('SettingWidget'):
            try:
                w.setDefault()
            except (Exception, ), exc:
                excs.append((w.plugin.Name, w.setting.Name, exc))
        self.settingExcDialog(excs, 'Exceptions During Set Default')
        self.enableApplyReset(True, True)

    def settingsReset(self):
        """ set each setting to its previous value

        @return None
        """
        excs = []
        for w in self.queryList('SettingWidget'):
            try:
                w.reset()
            except (Exception, ), exc:
                excs.append((w.plugin.Name, w.setting.Name, exc))
        self.settingExcDialog(excs, 'Exceptions During Reset')
        self.enableApplyReset(False, False)
        
    def settingApply(self):
        """ not final

        """
        excs = []
        for w in self.queryList('SettingWidget'):
            try:
                w.setFromValue()
            except (Exception, ), exc:
                excs.append((w.plugin.Name, w.setting.Name, exc))
        self.settingExcDialog(excs, 'Exceptions During Apply')
        self.enableApplyReset(False, False)
        self.emit(Signals.statusMessage, ('Saving settings...', ))
        self.emit(Signals.berylSettingChanged, ())

    def settingExcDialog(self, exceptions, caption):
        """ show an error dialog

        @param exceptions list of three-tuples
        @param caption dialog caption
        @return None
        """
        if not exceptions:
            return
        excstrs = QStringList()
        for exc in exceptions:
            excstrs.append('Plugin:%s Setting:%s Exception:%s' % exc)
        KMessageBox.errorList(None, caption, excstrs)


class SinglePage(Frame, SettingPage):
    """ SinglePage -> frame to display a single group of settings

    """
    def __init__(self, parent):
        Frame.__init__(self, parent, marginHint, spacingHint)
        self.header = ContentHeader(self)
        self.main = MultiSettingFrame(self)
        self.footer = ContentFooter(self)
        self.buildConnections()

    def buildConnections(self):
        """ build the connections for this instance

        @return None
        """
        SettingPage.buildConnections(self)
        self.connect(self.main.container, Signals.someChange,
                     self.onSomethingChanged)        

    def showSettings(self, plugin, arg):
        """ displays the settings page with individual setting frames

        @param plugin berylsettings Plugin instance
        @param arg setting name or setting section name
        @return None
        """
        if not self.unsavedCheck(plugin, arg):
            return
        self.header.setFromPlugin(plugin, arg)
        self.main.setPlugin(plugin, arg)


class TabPage(Frame, SettingPage):
    """  TabPage -> tab widget to display multiple groups of settings

    """
    def __init__(self, parent):
        Frame.__init__(self, parent, marginHint, spacingHint)
        self.header = ContentHeader(self)
        self.tabs = KTabWidget(self)
        sep = KSeparator(KSeparator.Horizontal, self)
        self.footer = ContentFooter(self)
        self.buildConnections()

    def setGroups(self, plugin):
        """ builds new page for each plugin setting group

        @param plugin berylsettings Plugin instance
        @return None
        """
        if not self.unsavedCheck(plugin, None):
            return
        self.header.setFromPlugin(plugin)
        self.clearPages()
        tabs = self.tabs
        changed = Signals.someChange        
        for group in plugin.groups:
            page = MultiSettingFrame(tabs)
            page.setPlugin(plugin, group.settings(plugin))
            tabs.addTab(page, i18n(group.label))
            self.connect(page.container, changed, self.onSomethingChanged)
        tabs.setCurrentPage(0)

    def clearPages(self):
        """ deletes all pages from the tab widget child

        @return None
        """
        tabs = self.tabs
        page = tabs.currentPage()
        while page:
            tabs.removePage(page)
            page.deleteLater()
            page = tabs.currentPage()            


class MultiSettingFrame(Frame):
    """ MultiSettingFrame -> a scrollable view of multiple settings

    """
    def __init__(self, parent, *args, **kwds):
        Frame.__init__(self, parent)
        self.frames = []
        self.previous = (None, None)        
        self.scroll = ContentScroll(self)
        viewport = self.scroll.viewport()
        self.container = Frame(viewport, marginHint, spacingHint, False)
        self.scroll.addChild(self.container)
        
    def setPlugin(self, plugin, arg):
        """ displays the settings page with individual setting frames

        @param plugin berylsettings Plugin instance
        @param arg setting name, setting type name, or sequence of settings
        @return None
        """
        if isinstance(arg, basestring):
            settings = plugin.settings[arg]
        elif isinstance(arg, Setting):
            settings = [arg, ]
        elif isinstance(arg, (list, tuple)):
            settings = arg
        else:
            print 'unknown setting type', arg
            settings = ()
        if self.previous == (plugin, settings):
            return
        self.previous = (plugin, settings)
        self.clearFrames()
        container = self.container
        layout = container.layout()
        changed = Signals.someChange
        frames = self.frames
        for setting in settings:
            frame = SettingFrame(container, plugin, setting)
            frame.show()
            frames.append(frame)            
            layout.addWidget(frame, 0, Qt.AlignTop)
            self.connect(frame, changed, container, changed)
        layout.addStretch(1)

    def clearFrames(self):
        """ removes child frames and clears the container frame

        @return None
        """
        for widget in self.frames:
           widget.close(True)
        self.container.layout().deleteAllItems()
        self.frames = []


class SettingFrame(QFrame):
    """ SettingFrame -> display and editing widgets for a single Setting.

    """
    def __init__(self, parent, plugin, setting):
        QFrame.__init__(self, parent)
        mainLayout = QVBoxLayout(self)
        innerLayout = QHBoxLayout(mainLayout)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.infoCaption = setting.ShortDesc
        self.infoText = setting.LongDesc
        self.infoButton = info = SmallPushButton(iconSet('help'), '', self)
        self.resetButton = reset = \
            SmallPushButton(iconSet('locationbar_erase'), '', self)
        QToolTip.add(reset, i18n('Reset to default value'))
        self.settingWidget = widget = settingWidget(self, plugin, setting)
        if isinstance(widget.layout(), QHBoxLayout):
            align = Qt.AlignLeft | Qt.AlignVCenter
        else:
            align = Qt.AlignLeft | Qt.AlignTop
        innerLayout.addWidget(info, 0, align)
        innerLayout.addWidget(reset, 0, align)
        innerLayout.addWidget(widget, 1)
        mainLayout.addSpacing(spacingHint)

        connect = self.connect
        connect(info, Signals.clicked, self.showTip)
        connect(reset, Signals.clicked, self.settingWidget.setDefault)
        connect(widget, Signals.someChange, self, Signals.someChange)

    def showTip(self):
        """ show a balloon tip with the setting description

        @return None
        """
        parent = self.resetButton
        pop = KPassivePopup.message(KPassivePopup.Balloon,
                                    self.infoCaption,
                                    self.infoText,
                                    icon('help', size=KIcon.SizeLarge),
                                    parent)
        pos = pop.pos()
        pos.setY(pos.y() + (parent.height() / 2))
        pos.setX(pos.x() + (parent.width() * 2))
        pop.show()
        pop.move(pos)        


class ContentHeader(QFrame):
    """ ContentHeader -> header frame with a pixmap and label for plugin

    """
    def __init__(self, parent):
        QFrame.__init__(self, parent)
        layout = QHBoxLayout(self)
        self.pluginIconLabel = iconLabel = QLabel(self)
        self.pluginNameLabel = nameLabel = QLabel(self)
        layout.addWidget(iconLabel)
        layout.addWidget(nameLabel, 1)

    def setFromPlugin(self, plugin, extra=''):
        """ sets this header icon and label to plugin values

        @param plugin berylsettings Plugin instance
        @param extra any object; if string or unicode, added to end of label
        @return None
        """
        ico = plugin.icon(KIcon.SizeLarge, iconLoader())
        self.pluginIconLabel.setPixmap(ico)
        if isinstance(extra, basestring) and extra:
            extra = ' - %s' % (extra, )
        else:
            extra = ''
        values = (plugin.ShortDesc, extra)
        self.pluginNameLabel.setText('<b>%s%s</b>' % values)


class ContentFooter(QFrame):
    """ ContentFooter -> footer frame with common buttons

    """
    def __init__(self, parent):
        QFrame.__init__(self, parent)
        layout = QHBoxLayout(self)
        self.helpButton = guiButton(self, 'help', layout)
        self.defaultsButton = guiButton(self, 'defaults', layout)
        layout.addStretch(1)
        self.applyButton = guiButton(self, 'apply', layout, False)
        self.resetButton = guiButton(self, 'reset', layout, False)


class ContentScroll(QScrollView):
    """ ContentScroll -> a scroll view fitted to a single child

    The redefined sizeHint method is important -- without it, the
    widget size isn't picked up correctly by the parent.  Found in
    kcontrol sources.
    """
    def __init__(self, parent):
        QScrollView.__init__(self, parent)
        self.setResizePolicy(QScrollView.AutoOneFit)
        self.setFrameShape(self.NoFrame)

    def sizeHint(self):
        """ recommended size for this widget

        @return QSize instance
        """
        return self.minimumSizeHint()
