#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" kberylsettings.lib -> library classes, functions, and attributes.

"""
from qt import PYSIGNAL, QPixmap, QImage, SIGNAL, SLOT
from kdecore import KGlobal, KIcon, KShortcut, i18n
from kdeui import KAction, KStdAction


##
# a cache for already-created icons
iconCache = {}


class Signals:
    """ class with SIGNAL and PYSIGNAL attributes

    """
    actionStatusText = SIGNAL('actionStatusText(const QString &)')
    activated = SIGNAL('activated()')
    activatedIndex = SIGNAL('activated(int)')
    berylContextChanged = PYSIGNAL('berylContextChanged')
    berylPluginEnabled = PYSIGNAL('berylPluginEnabled')
    berylSettingChanged = PYSIGNAL('berylSettingChanged')
    capturedShortcut = SIGNAL('capturedShortcut (const KShortcut &)')
    changed = SIGNAL('changed()')
    clearStatusText = SIGNAL('clearStatusText()')
    clicked = SIGNAL('clicked()')
    clickedId = SIGNAL('clicked(int)')
    colorChanged = SIGNAL('changed(const QColor &)')
    doubleValueChanged = SIGNAL('valueChanged(double)')
    iconSizeChanged = PYSIGNAL('iconSizeChanged')
    intValueChanged = SIGNAL('valueChanged(int)')
    itemClicked = SIGNAL('clicked(QListViewItem *)')
    itemSelected = SIGNAL('selectionChanged(QListViewItem *)')
    okClicked = SIGNAL('okClicked()')
    partURL = PYSIGNAL('partURL')
    quitSelected = SIGNAL('quitSelected()')
    searchInput = PYSIGNAL('searchInput')
    selectItem = PYSIGNAL('selectItem')
    selectPrevious = PYSIGNAL('selectPrevious')
    showAbout = PYSIGNAL('showAbout')
    showGroups = PYSIGNAL('showGroups')
    showSettings = PYSIGNAL('showSettings')
    someChange = PYSIGNAL('someChange')
    statusMessage = PYSIGNAL('statusMessage')
    textChanged = SIGNAL('textChanged(const QString&)')
    toggled = SIGNAL('toggled(bool)')
    viewModeChanged = PYSIGNAL('viewModeChanged')    


class Slots:
    """ class with SLOT attributes

    """
    captureShortcut = SLOT('captureShortcut()')
    clear = SLOT('clear()')
    messageText = SLOT('message( const QString & )')


def iconLoader():
    """ returns the global icon loader

    @return KIconLoader instance
    """
    return KGlobal.instance().iconLoader()


def icon(name, group=KIcon.NoGroup, size=KIcon.SizeSmall):
    """ loads pixmap by name

    This function could return the icon directly from the loadIcon
    call, but that method (for some reason) doesn't apply any icon
    effects.

    @param name icon name
    @param group KIcon group
    @param size KIcon size
    @return icon as QPixmap instance
    """
    try:
        return iconCache[(name, size)]
    except (KeyError, ):
        pass
    ico = iconLoader().loadIcon(name, group)
    img = ico.convertToImage()
    pix = iconCache[(name, size)] = QPixmap()
    pix.convertFromImage(img.smoothScale(size, size, QImage.ScaleMin))
    return pix


def iconSet(name, group=KIcon.NoGroup, size=KIcon.SizeSmall):
    """ load icon set by name
    
    @param name icon set name
    @param group KIcon group
    @param size KIcon size
    @return icon as QIconSet instance
    """
    return iconLoader().loadIconSet(name, group, size)


def action(text, iconset, cut, slot, collection, name,
           tooltip='', pluggable=None):
    """ creates and configures a KAction instance

    @param text display value for action
    @param iconset name of icon set for action
    @param cut
    @param slot
    @param collection KActionCollection instance
    @param name internal action name
    @keyparam tooltip='' value for tooltip
    @keyparam pluggable=None any pluggable object
    @return KAction instance
    """
    if not cut:
        cut = KShortcut()
    else:
        cut = KShortcut(cut)
    iconset = iconSet(iconset)
    action = KAction(text, iconset, cut, slot, collection, name)
    if tooltip:
        action.setToolTip(tooltip)
    if pluggable:
        action.plug(pluggable)
    return action


def stdAction(attr, slot, collection, name, pluggable=None):
    """ creates a named KStdAction instance

    @param attr attribute of KStdAction class
    @param slot slot for KStdAction instance
    @param collection KActionCollection instance
    @param name internal action name
    @keyparam pluggable=None any pluggable object    
    @return KStdAction instance
    """
    typ = getattr(KStdAction, attr)
    action = typ(slot, collection, name)
    if pluggable:
        action.plug(pluggable)
    return action
