#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" kberylsettings.main -> defines KBerylSettings main window.

"""
from os import listdir
from os.path import abspath, dirname, expanduser
from sys import argv

from kdecore import KApplication, KCmdLineArgs, KIcon, KGlobal, KWin, i18n, locate
from kdeui import KMainWindow, KStdGuiItem, KLineEdit, KSystemTray, KTipDialog, \
     KKeyDialog
from kfile import KFileDialog
from qt import Qt, QWidget, QHBoxLayout, QLabel, QImage

from kberylsettings.aboutapp import about
from kberylsettings.beryl import Context
from kberylsettings.contentframe import ContentFrame
from kberylsettings.lib import Signals, Slots, action, stdAction, icon
from kberylsettings.plugindialog import PluginDialog
from kberylsettings.pluginframe import PluginFrame
from kberylsettings.widget import Frame, Popup, SmallPushButton, Splitter


class KBerylSettings(KMainWindow):
    """ The main window.

    """
    viewMode = PluginFrame.categoryPageId
    iconSize = KIcon.SizeMedium
    startDir = '::kberylsettings'

    def __init__(self):
        KMainWindow.__init__(self)
        self.buildWidgets()
        self.buildMenus()
        self.buildSystray()
        self.buildConnections()
        self.buildFinal()
        
    def buildWidgets(self):
        """ builds widgets on this instance

        @return None
        """
        actions = self.actionCollection()
        actions.setHighlightingEnabled(True)
        self.mainSplitter = splitter = Splitter(self, Qt.Horizontal)
        self.leftFrame = left = Frame(splitter)
        self.rightFrame = right = Frame(splitter)
        self.contentFrame = ContentFrame(right)
        self.searchWidget = search = QWidget(self.leftFrame)
        searchLayout = QHBoxLayout(search, 6, 6)
        searchLayout.setAutoAdd(True)
        clear = KStdGuiItem.clear()
        self.clearSearch = SmallPushButton(clear.iconSet(), '', search)
        self.labelSearch = QLabel(i18n('Search: '), search)
        self.inputSearch = KLineEdit(search)
        self.pluginList = PluginFrame(left)
        self.setCentralWidget(splitter)
        statusBar = self.statusBar()

    def buildMenus(self):
        """ builds the menus for this instance

        @return None
        """
        menu = self.menuBar()
        menu.insertItem(i18n('&File'), self.fileMenu())
        menu.insertItem(i18n('&View'), self.viewMenu())
        menu.insertItem(i18n('&Settings'), self.settingsMenu())        
        menu.insertItem(i18n('&Help'), self.helpMenu())

    def buildSystray(self):
        """ builds the system tray icon for this instance

        @return None
        """
        self.sysTray = KSystemTray(self)
        pixmap = icon(self.iconFile, size=KIcon.SizeSmallMedium)
        self.sysTray.setPixmap(pixmap)
        pop = self.sysTray.contextMenu()
        self.pluginsAction.plug(pop)
        cap = KApplication.kApplication().caption()
        pop.changeTitle(pop.idAt(0), pixmap, cap)
        pop.insertItem(i18n('Named &Profiles'), self.profileMenu())
        self.sysTray.show()

    def iconFile(self):
        """ names the default icon file

        @return icon filename 
        """
        return locate('icon', 'kberylsettings.svg')
    iconFile = property(iconFile)
    
    def buildConnections(self):
        """ builds the connections for this instance

        @return None
        """
        connect = self.connect
        connect(self, Signals.viewModeChanged, self.onMode)        
        connect(self, Signals.iconSizeChanged, self.onIconSize)
        connect(self.clearSearch, Signals.clicked, self.onClearSearch)        
        connect(self.inputSearch, Signals.textChanged, self.onInputSearch)
        connect(self.sysTray, Signals.quitSelected, self.close)
        statusBar = self.statusBar()
        actions = self.actionCollection()
        connect(actions, Signals.actionStatusText, self.statusBar(),
                Slots.messageText)
        connect(actions, Signals.clearStatusText, self.statusBar(),
                Slots.clear)

    def buildFinal(self):
        """ do everything else needed to create the widget

        @return None
        """
        self.context = context = Context()
        self.emit(Signals.berylContextChanged, (context, ))
        connect = self.connect
        connect(context, Signals.statusMessage, self.showMessage)
        self.setIcon(icon(self.iconFile, size=KIcon.SizeSmall))
        KWin.setIcons(self.winId(),
                      icon(self.iconFile, size=KIcon.SizeLarge),
                      icon(self.iconFile, size=KIcon.SizeSmall))
        config = self.config('mainwindow')
        self.restoreWindowSize(config)
        self.mainSplitter.setSizes([self.width()*0.25, self.width()*0.75])
        self.onIntro()
        config = self.config('mainview')
        emit = self.emit
        emit(Signals.iconSizeChanged,
             (config.readNumEntry('iconsize', self.iconSize), ))
        viewMode = config.readNumEntry('viewmode', self.viewMode)
        emit(Signals.viewModeChanged, (viewMode, ))
        actions = self.actionCollection()
        actions.readShortcutSettings('shortcuts', self.config())
        ## done after initial emit to prevent spurious context update
        connect(self, Signals.berylContextChanged, self.onContextChanged)
        self.tipDialog(force=False)
        
    def config(self, group=None):
        """ loads the general configuration object

        @keyparam group=None if provided, config group set to this value
        @return KConfig instance
        """
        config = KGlobal.instance().config()
        if group is not None:
            config.setGroup(group)
        return config

    def fileMenu(self):
        """ creates a new File menu

        @return KPopupMenu instance
        """
        pop = Popup(self)
        actions =  self.actionCollection()
        self.importAction = action('Import settings...', 'fileopen', 'Ctrl+O',
                                   self.onImport, actions, 'import_action',
                                   'Import settings from a file.',
                                   pop)
        self.exportAction = action('Export settings...', 'filesave', 'Ctrl+S',
                                   self.onExport, actions, 'export_action',
                                   'Export settings to a file.',
                                   pop)
        pop.insertSeparator()
        pop.insertItem(i18n('Named &Profiles'), self.profileMenu())
        self.quitAction = stdAction('quit', self.close, actions, '', pop)
        return pop

    def helpMenu(self):
        """ creates a new Help menu

        @return KPopupMenu instance
        """
        actions = self.actionCollection()        
        pop = KMainWindow.helpMenu(self)
        self.tipAction = stdAction('tipOfDay', self.tipDialog, actions,
                                   'tipOfDay', None)
        self.introAction = action('KDE Beryl Settings Introduction',
                                  'help', 'Ctrl+I', self.onIntro,
                                  actions, 'intro_action',
                                  'Show the introduction.', None)
        pop.insertSeparator(2)
        self.tipAction.plug(pop, 3)
        self.introAction.plug(pop, 4)
        return pop

    def profileMenu(self):
        """ creates a new Named Profiles menu

        """
        try:
            profileMap = self.profileMap
        except (AttributeError, ):
            profileMap = self.profileMap = {}
            
        pop = Popup(self)
        names = [n for n in listdir(expanduser('~/.beryl'))
                     if n.endswith('.Profile')]
        names.sort()
        for name in names:
            popid = pop.insertItem(name)
            index = hash(name)
            profileMap[index] = name
            pop.setItemParameter(popid, index)
            pop.connectItem(popid, self.onProfileSelect)
        return pop
            
    def viewMenu(self):
        """ creates a new View menu

        @return KPopupMenu instance
        """
        pop = Popup(self)
        pop.insertItem(i18n('&Mode'), self.modeMenu())
        pop.insertItem(i18n('Icon &Size'), self.iconSizeMenu())
        return pop

    def settingsMenu(self):
        """ creates a new Settings menu

        @return KPopupMenu instance
        """
        actions = self.actionCollection()
        pop = Popup(self)
        self.keysAction = stdAction('keyBindings', self.onConfigKeys,
                                    actions, 'keyBindings', pop)
        self.pluginsAction = action('Select Plugins...', 'configure', 'Ctrl+P',
                                    self.onSelectPlugins,
                                    actions,
                                    'select_plugins',
                                    'Select Beryl plugins.',
                                    pop)
        return pop

    def modeMenu(self):
        """ creates a new View Mode menu

        @return KPopupMenu instance
        """
        pop = self.modePopup = Popup(self)
        ids = self.modePopupIds = {}
        for value, label in PluginFrame.modePages:
            modeid = ids[value] = pop.insertItem(i18n(label))
            pop.setItemParameter(modeid, value)
            pop.connectItem(modeid, self, Signals.viewModeChanged)
        return pop

    def iconSizeMenu(self):
        """ creates a new View Icon Size menu

        @return KPopupMenu instance
        """
        pop = self.iconSizePopup = Popup(self)
        ids = self.iconSizePopupIds = {}
        for size, label in iconSizes():
            sizeid = ids[size] = pop.insertItem(i18n(label))
            pop.setItemParameter(sizeid, size)
            pop.connectItem(sizeid, self, Signals.iconSizeChanged)
        pop.setItemChecked(sizeid, True)
        return pop

    def queryClose(self):
        """ abuse the queryClose method because other close methods are goofy

        This method may grow in the future to prompt the user for
        unsaved changes.
        
        @return True
        """
        if not self.sender():
            self.hide()
            return False
        if not self.contentFrame.queryClose():
            return False
        self.saveMainWindowSettings(self.config())
        self.saveWindowSize(self.config('mainwindow'))
        config = self.config('mainview')
        config.writeEntry('iconsize', self.iconSize)
        config.writeEntry('viewmode', self.viewMode)
        actions = self.actionCollection()
        actions.writeShortcutSettings('shortcuts', config)
        return True

    def onClearSearch(self):
        """ clears the search input, emits signal to set view mode

        @return None
        """
        self.inputSearch.clear()
        self.emit(Signals.viewModeChanged, (self.viewMode, ))

    def onConfigKeys(self):
        """ displays the kde shortcut key config dialog for our actions

        @return None
        """
        KKeyDialog.configure(self.actionCollection(), self)

    def onContextChanged(self):
        """ writes the beryl context and sends a reload message

        @return None
        """
        self.context.write()

    def onIconSize(self, value):
        """ changes the View Icon Size menu

        @param value new icon size
        @return None
        """
        self.iconSize = value
        for size, menuid in self.iconSizePopupIds.items():
            self.iconSizePopup.setItemChecked(menuid, size==value)

    def onInputSearch(self, value):
        """ emits signal to change view mode and signal to search for text

        """
        self.emit(Signals.searchInput, (value, ))
        self.emit(Signals.viewModeChanged, (PluginFrame.searchPageId, ))

    def onIntro(self):
        """ emits signal to show the introduction

        """
        self.emit(Signals.showAbout, (self.context, None))        

    def onPluginDialogOk(self):
        """ sets current context active plugins

        @return None
        """
        items = self.pluginDialogItems.items()
        active = [name for name, enabled in items if enabled]
        context = self.context
        for p in context.plugins:
            self.emit(Signals.berylPluginEnabled, (p, p.Name in active))
        context.active = active
        self.emit(Signals.berylContextChanged, (context, ))

    def onProfileSelect(self, itemId):
        """ sets current profile

        @param itemId profile popup menu item id
        @return None
        """
        self.context.Profile = self.profileMap[itemId]
        self.emit(Signals.berylContextChanged, (self.context, ))
        
    def onSelectPlugins(self):
        """ shows dialog to select/deselect active plugins

        @return None
        """
        self.pluginDialogItems = {}
        dialog =  PluginDialog(self, self.context, self.pluginDialogItems)
        self.connect(dialog, Signals.okClicked, self.onPluginDialogOk)
        dialog.exec_loop()

    def onMode(self, value):
        """ changes the View Mode menu, enables/disables Icon Size menu items

        @param value new view mode
        @return None
        """
        if value in PluginFrame.transitoryModes:
            return
        self.viewMode = value
        iconMode = value in PluginFrame.iconSizeModes
        for menuid in self.iconSizePopupIds.values():
            self.iconSizePopup.setItemEnabled(menuid, iconMode)
        for mode, menuid in self.modePopupIds.items():
            self.modePopup.setItemChecked(menuid, mode==value)
            
    def onImport(self):
        """ shows dialog to select profile for import

        @return None
        """
        filename = KFileDialog.getOpenFileName(self.startDir, '', None,
                                               'Select Profile to Import')
        if filename:
            self.context.Import(filename, True) # overwrite
            self.emit(Signals.berylContextChanged, (self.context, ))
            self.showMessage('Settings imported.')

    def onExport(self):
        """ shows dialog to select filename for profile export

        @return None
        """
        filename = KFileDialog.getSaveFileName(self.startDir, '', None,
                                               'Select File to Export Profile')
        if filename:
            print 'exporting file...', filename
            
    def showMessage(self, text):
        """ show a message in the status bar

        @param text value to show
        @return None
        """
        self.statusBar().message(text, 3000)

    def tipDialog(self, force=True):
        """ show a KTipDialog

        @keyparam force=True if true, forces dialog display.  default
        is True for easy connection to the menu entry.
        @return None
        """
        KTipDialog.showTip(self, locate('data', 'tips'), force)


def iconSizes():
    """ creates list of KIcon sizes and their names

    @return sequence of two-tuples in the form of (size, name)
    """
    def sizeOk(attr):
        return attr.startswith('Size') and attr != 'SizeSmallMedium'
    names = [name for name in dir(KIcon) if sizeOk(name)]
    sizes = [getattr(KIcon, name) for name in names]
    names = [name.replace('Size', '') for name in names]
    sizenames = zip(sizes, names)
    sizenames.sort()
    return sizenames


def main():
    """ runs the KBerylSettings window with a KApplication

    @return None
    """
    aboutdata = about()
    KCmdLineArgs.init(argv, aboutdata)
    app = KApplication()
    dirs = app.dirs()
    basedir = abspath(dirname(__file__))    
    dirs.addResourceDir('data', basedir)
    dirs.addResourceDir('templates', basedir + '/html')    
    dirs.addResourceDir('icon', basedir + '/pixmaps')
    aboutdata.setProgramLogo(QImage(locate('icon', 'kberylsettings.png')))
    obj = KBerylSettings()    
    obj.show()
    app.exec_loop()


if __name__ == '__main__':
    main()
