#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" kberylsettings.pluginframe -> defines the left side plugin views
and search view.

"""
from re import sub

from qt import QLabel, QListViewItem
from kdecore import KIcon, i18n

from kberylsettings.beryl import Setting
from kberylsettings.lib import Signals, icon, iconLoader
from kberylsettings.widget import BasicListView, Frame, ValueListViewItem, \
     WidgetStack


class PluginFrame(WidgetStack):
    """ PluginFrame -> stack with various plugin views and search widget.

    """
    iconPageId, treePageId, searchPageId, categoryPageId = pageIds = range(4)
    iconSizeModes = (iconPageId, categoryPageId)
    transitoryModes = (searchPageId, )
    modePages = (
        (iconPageId, '&Icon View'),
        (treePageId, '&Tree View'),
        ## search does not have a mode menu
        ## entry selectable by the user.
        (categoryPageId, '&Category View')
    )
    relatedModes = {
        iconPageId : (treePageId, categoryPageId),
        treePageId : (iconPageId, categoryPageId),
        categoryPageId : (iconPageId, treePageId),
    }
    
    def __init__(self, parent):
        WidgetStack.__init__(self, parent)
        pages = (IconView(self), TreeView(self), SearchView(self),
                 CategoryView(self))
        addWidget = self.addWidget
        for pageId, page in zip(self.pageIds, pages):
            addWidget(page, pageId)
        connect = self.connect
        root = self.topLevelWidget()        
        connect(root, Signals.viewModeChanged, self.raiseWidget)

        for pageId, others in self.relatedModes.items():
            for otherId in others:
                connect(pages[pageId], Signals.selectItem,
                        pages[otherId].otherSelected)
        
        for page in pages:
            connect(page, Signals.berylSettingChanged, root.onContextChanged)
            connect(page, Signals.showAbout, root, Signals.showAbout)
            connect(page, Signals.showSettings, root, Signals.showSettings)            
            connect(page, Signals.showGroups, root, Signals.showGroups)
            connect(page, Signals.statusMessage, root.showMessage)
            try:
                connect(root, Signals.berylContextChanged, page.onBerylContext)
                connect(root, Signals.berylPluginEnabled, page.enablePluginItem)
                connect(root, Signals.selectPrevious, page.selectPluginItem)
            except (AttributeError, ):
                pass


class SearchView(Frame):
    """ SearchView -> a frame to display found keywords and their results

    """
    def __init__(self, parent):
        Frame.__init__(self, parent)
        self.index = {}
        self.loader = iconLoader()
        self.keywordLabel = QLabel(i18n('Keywords:'), self)
        self.keywordsList = keywordsList = BasicListView(self)
        self.resultsLabel = QLabel(i18n('Results:'), self)
        self.resultsList = resultsList = BasicListView(self)
        connect = self.connect
        connect(keywordsList, Signals.itemSelected, self.showResults)
        connect(resultsList, Signals.itemSelected, self.emitResult)
        root = self.topLevelWidget()
        connect(root, Signals.searchInput, self.showKeywords)
        connect(root, Signals.berylContextChanged, self.rebuildIndex)

    def rebuildIndex(self, context):
        """ reconstructs the search 'index' given a berylsetting context

        This might look slow at first glance, but my machine, this
        function takes less than 0.1 second.

        @param context berylsetting.Context instance
        @return None
        """
        self.context = context
        index = self.index
        index.clear()
        itemKeywords = self.itemKeywords
        for p in context.plugins[::-1]:
            for key in itemKeywords(p):
                seq = index.setdefault(key, [])
                seq.append((p, None))
            for s in p.Settings:
                for key in itemKeywords(s):
                    seq = index.setdefault(key, [])
                    seq.append((p, s))

    def itemKeywords(self, obj):
        """ creates a list of keywords for the given berylsettings object

        @param obj berylsettings Plugin or Setting instance
        @return list of keywords
        """
        values = [obj.ShortDesc, obj.LongDesc, ] + \
                     obj.ShortDesc.split() + \
                     obj.LongDesc.split()
        return [sub('\W|_', ' ', v) for v in values]

    def showKeywords(self, text):
        """ displays keywords from the index that match the given text

        @param text search value
        @return None
        """
        text = str(text)        
        if not text:
            return
        searchtext = text.lower()
        index = self.index
        parent = self.keywordsList
        parent.clear()
        active = self.context.active
        for keyword in self.index:
            if searchtext in keyword.lower():
                plugin, setting = index[keyword][0]
                if plugin.Name in active:
                    item = ValueListViewItem(parent, keyword, index[keyword])

    def showResults(self, item):
        """ displays plugin/setting items from the selected keyword

        @return None
        """
        parent = self.resultsList
        parent.clear()
        cache = {}
        for plugin, setting in item.value:
            desc = plugin.ShortDesc
            item = cache.get(plugin)
            if item is None:
                value = (plugin, None)
                item = cache[plugin] = ValueListViewItem(parent, desc, value)
                item.setPixmap(0, plugin.icon(KIcon.SizeSmall, self.loader))
            if setting:
                desc = setting.ShortDesc
                sub = cache.get((plugin, setting))
                if sub is None:
                    sub = cache[(plugin, setting)] = \
                          ValueListViewItem(item, desc, (plugin, setting))
            item.setOpen(True)

    def emitResult(self, item):
        """ emits signal to show a plugin or a setting

        @param item result list view item
        @return None
        """
        plugin, setting = item.value
        if setting is None:
            self.emit(Signals.showAbout, (self.context, plugin, ))
        else:
            self.emit(Signals.showSettings, (plugin, Setting(setting), ))


class BasicPluginView(BasicListView):
    """ BasicPluginView -> mixin for shared view functionality

    """
    def __init__(self, parent):
        BasicListView.__init__(self, parent)
        self.loader = iconLoader()
        self.connect(self, Signals.itemClicked, self.onItemClick)

    def selectPluginItem(self, plugin, setting, activate=False):
        """ locate and set current item given plugin and setting

        @param plugin beryl settings Plugin object
        @param setting any object
        @return None
        """
        def match(a):
            return a.plugin and \
                   a.plugin.Name == plugin.Name and \
                   a.setting == setting
        for item in self:
            if match(item):
                self.setCurrentEnsureVisible(item)
                if activate:
                    self.onItemClick(item)
                break
        
    def addItems(self):
        """ adds list view items for every plugin in context

        @return None
        """
        active = self.context.active
        for plugin in self.context.plugins[::-1]:
            self.listItem(plugin, active)

    def listItem(self, plugin, active):
        """ create a list view item for a plugin

        @param plugin beryl settings Plugin object
        @param active sequence of active plugin names
        @return QListViewItem instance
        """
        item = None
        for i in self:
            if i.plugin and i.plugin.Name == plugin.Name and not i.setting:
                item = i
                break
        if not item:
            item = ViewItem(self, plugin.ShortDesc)
        item.plugin = plugin
        item.setVisible(plugin.Name in active)
        item.setPixmap(0, plugin.icon(self.iconSize, self.loader))
        return item

    def onBerylContext(self, context):
        """ beryl settings context has changed

        @param context
        @return None
        """
        self.context = context
        self.addItems()

    def enablePluginItem(self, plugin, enable):
        """ toggle the enabled state of a plugin list view item

        @param index position of plugin within plugins list
        @return None
        """
        for item in self:
            if item.plugin == plugin and not item.setting:
                item.setVisible(enable)

    def setCurrentEnsureVisible(self, item):
        """ set current item and ensure it's visible

        @param QListViewItem instance
        @return None
        """
        self.setCurrentItem(item)
        self.ensureItemVisible(item)


class IconView(BasicPluginView):
    """ IconView -> icon list view of plugins

    """
    iconSize = KIcon.SizeMedium

    def __init__(self, parent):
        BasicPluginView.__init__(self, parent)
        root = self.topLevelWidget()        
        self.connect(root, Signals.iconSizeChanged, self.onIconSize)
        
    def onIconSize(self, size):
        """ update the icon size

        @param size icon size as integer
        @return None
        """
        self.iconSize = size
        for item in self:
            if item.setting or not item.plugin:
                item.setPixmap(0, Setting.icon(item.setting, size))                
            else:
                item.setPixmap(0, item.plugin.icon(size, self.loader))

    def onItemClick(self, item):
        """ emit a signal if a plugin item is selected

        @param item QListViewItem instance
        @return None
        """
        if not item:
            return
        if item.plugin and not item.setting:
            self.showSettingItems(item.plugin)
            if self.sender() is self or self.isVisible():
                self.emit(Signals.showAbout, (self.context, item.plugin, ))
        elif item == self.firstChild():
            self.showPlugins()
        else:
            self.emit(Signals.showSettings, (item.plugin, item.setting))
        self.emit(Signals.selectItem, (item, ))

    def showPlugins(self):
        """ sets all active plugin items visible

        @return None
        """
        active = self.context.active
        for item in self:
            if item.plugin and item.plugin.Name in active and not item.setting:
                item.setVisible(True)
            else:
                item.setVisible(False)

    def backItem(self):
        """ locates or creates the 'Back' item

        @return ViewItem instance
        """
        back = self.findItem(i18n('Back'), 0)        
        if back is None:
            back = ViewItem(self, i18n('Back'))
            back.setPixmap(0, icon('back', size=self.iconSize))
        return back
    
    def showSettingItems(self, plugin):
        """ displays setting icons for given plugin

        @param plugin beryl settings Plugin object        
        @return None
        """
        keys = plugin.settings.keys()
        keys.sort()
        back = self.backItem()
        back.setVisible(True)
        settings = []
        for item in self:
            p, s = item.plugin, item.setting
            if p and s:
                if s in keys:
                    item.setVisible(True)
                    item.plugin = plugin
                    item.setting = s
                    settings.append(s)
                else:
                    item.setVisible(False)
            elif p and not s:
                item.setVisible(False)
        for key in keys[::-1]:
            if key not in settings:
                sub = ViewItem(self, key, plugin=plugin, setting=key)
                sub.setPixmap(0, Setting.icon(key, self.iconSize))
                sub.moveItem(back)                

    def otherSelected(self, other):
        """ set current item when other item is selected elsewhere

        @param other ViewItem instance
        @return None
        """
        if other.plugin and other.setting:
            self.showSettingItems(other.plugin)            
            for i in self:
                if i.plugin and \
                       i.plugin.Name == other.plugin.Name and \
                       i.setting == other.setting:
                   self.setCurrentEnsureVisible(i)
                   break
        elif other.plugin:
            self.showPlugins()
            for i in self:
                if i.plugin and i.plugin.Name == other.plugin.Name:
                    self.setCurrentEnsureVisible(i)


class TreeView(BasicPluginView):
    """ TreeView -> tree list view of plugins

    """
    iconSize = KIcon.SizeSmall

    def __init__(self, parent):
        BasicPluginView.__init__(self, parent)
        self.setRootIsDecorated(True)        
        
    def listItem(self, plugin, active):
        """ create an item for the plugin and sub items for its settings

        @param plugin beryl settings Plugin object
        @param active sequence of active plugin names
        @param size icon size
        @return QListViewItem instance with children
        """
        item = BasicPluginView.listItem(self, plugin, active)
        groups = plugin.settings
        keys = groups.keys()
        keys.sort()
        child = item.firstChild()
        while child:
            item.takeItem(child)
            child = item.firstChild()
        for key in keys[::-1]:
            sub = ViewItem(item, key, plugin=plugin, setting=key)
            sub.setPixmap(0, Setting.icon(key, self.iconSize))
        return item

    def onItemClick(self, item):
        """ emit a signal if a plugin item is selected

        @param item QListViewItem instance
        @return None
        """
        if not item:
            return
        plugin, setting = item.plugin, item.setting
        if plugin and setting:
            self.emit(Signals.showSettings, (plugin, setting))            
        elif self.sender() is self or self.isVisible():
            self.emit(Signals.showAbout, (self.context, plugin, ))            
        self.emit(Signals.selectItem, (item, ))

    def otherSelected(self, other):
        """ set current item when other item is selected elsewhere

        @param other ViewItem instance
        @return None
        """
        if other.category:
            for i in self:
                if i.plugin == other.plugin and not i.setting:
                    self.setCurrentEnsureVisible(i)
                    break
        elif other.plugin and not other.setting:
            for i in self:
                if i.plugin == other.plugin and not i.setting:
                    self.setCurrentEnsureVisible(i)
                    break
        elif other.plugin and other.setting:
            for i in self:
                if i.plugin == other.plugin and i.setting == other.setting:
                    i.parent().setOpen(True)
                    self.setCurrentEnsureVisible(i)
                    break
            

class CategoryView(BasicPluginView):
    """ CategoryView -> tree list view of plugins by category

    """
    iconSize = KIcon.SizeMedium

    def __init__(self, parent):
        BasicPluginView.__init__(self, parent)
        self.setRootIsDecorated(True)
        root = self.topLevelWidget()        
        self.connect(root, Signals.iconSizeChanged, self.onIconSize)
        
    def onIconSize(self, size):
        """ update the icon size

        @param size icon size as integer
        @return None
        """
        self.iconSize = size
        for i in self:
            if i.category:
                i.setPixmap(0, i.category.icon(size))
            elif i.plugin:
                i.setPixmap(0, i.plugin.icon(size, self.loader))

    def onItemClick(self, item):
        """ emit appropriate signal when plugin or category selected

        @param item QListViewItem instance
        @return None
        """
        if not item:
            return
        if item.plugin:
            self.emit(Signals.showGroups, (item.plugin, ))
            self.emit(Signals.selectItem, (item, ))

    def addItems(self):
        """ adds list view items for each category and plugin in context

        @return None
        """
        active = self.context.active
        seen = []
        for category in self.context.categories[::-1]:
            plugins = [p for p in category.plugins[::-1] if p.Name in active]
            if not category.Name:
                ## general plugin(s)
                for plugin in category.plugins:
                    if plugin.Name in active:
                        item = self.pluginItem(plugin, self)
                        seen.append(item)
            else:
                catitem = self.categoryItem(category)
                catitem.setVisible(bool(plugins))
                seen.append(catitem)
                for plugin in plugins:
                    if plugin.Name in active:
                        item = self.pluginItem(plugin, catitem)
                        seen.append(item)
        for item in self:
            if item.isVisible():
                item.setVisible(item in seen)
        self.setSorting(1)                        
        self.sort()
        self.setSorting(-1)
        
    def categoryItem(self, category):
        """ create a list view item for a category

        @param category
        @return QListViewItem instance
        """
        desc = category.ShortDesc
        item = self.findItem(desc, 0)
        if not item:
            item = CategoryViewItem(self, desc, category=category)
        item.setPixmap(0, category.icon(self.iconSize))
        return item

    def pluginItem(self, plugin, parent):
        """ create a list view item for a plugin

        @param category
        @return QListViewItem instance
        """
        for item in [i for i in self if i.parent() in (parent, 0)]:
            if item.plugin.Name == plugin.Name:
                item.plugin = plugin
                return item
        if parent is self:
            item = self.findItem(plugin.ShortDesc, 0)
            if item:
                item.plugin = plugin
                return item
        item = CategoryViewItem(parent, plugin.ShortDesc, plugin=plugin)
        item.setPixmap(0, plugin.icon(self.iconSize, self.loader))
        return item

    def otherSelected(self, other):
        """ set current item when other item is selected elsewhere

        @param other ViewItem instance
        @return None
        """
        for i in self:
            if i.plugin and i.plugin == other.plugin and i.parent():
                i.parent().setOpen(True)
                self.setCurrentEnsureVisible(i)                
                break


class ViewItem(QListViewItem):
    """ ViewItem -> QListViewItem with attributes common to local view classes

    """
    def __init__(self, parent, text, category=None, plugin=None, setting=None):
        QListViewItem.__init__(self, parent, text)
        self.category = category        
        self.plugin = plugin
        self.setting = setting


class CategoryViewItem(ViewItem):
    """ Sortable items for the category list view

    """
    def compare(self, other, column, ascending):
        """ compare this item to another

        @param other QListViewItem instance to compare against
        @param column ignored
        @param ascending ignored
        @return -1 if this item is less than other, 1 if greater, 0 if equal
        """
        for (obj, val) in ((self, -1), (other, 1)):
            if obj.plugin and obj.plugin.isGeneral:
                return val
        return cmp(self.text(0), other.text(0))
