#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" kberylsettings.widget -> various common widgets.

"""
from qt import Qt, QFrame, QGroupBox, QListViewItem, QListViewItemIterator, \
     QSizePolicy, QSplitter, QVBoxLayout, QWidgetStack
from kdeui import KListView, KPopupMenu, KPushButton, KStdGuiItem


class Frame(QFrame):
    """ Frame -> QFrame with QVBoxLayout

    """
    def __init__(self, parent, margin=0, spacing=-1, autoAdd=True):
        QFrame.__init__(self, parent)
        layout = QVBoxLayout(self, margin, spacing)
        layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        layout.setAutoAdd(autoAdd)


class Popup(KPopupMenu):
    """ Popup -> KPopupMenu always set to checkable

    """
    def __init__(self, parent):
        KPopupMenu.__init__(self, parent)
        self.setCheckable(True)


class Splitter(QSplitter):
    """ Splitter -> QSplitter always set to opaque resize

    """
    def __init__(self, parent, style):
        QSplitter.__init__(self, style, parent)
        self.setOpaqueResize(True)
        self.setChildrenCollapsible(False)


class SmallPushButton(KPushButton):
    """ SmallPushButton -> KPushButton with fixed sizes

    """
    def __init__(self, icon, text, parent):
        KPushButton.__init__(self, icon, text, parent)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))        


class ValueListViewItem(QListViewItem):
    """ ValueListViewItem -> QListViewItem with extra 'value' attribute

    """
    def __init__(self, parent, text, value=None):
        QListViewItem.__init__(self, parent, text)
        self.value = value


class ListView(KListView):
    """ ListView -> KListView that is directly iterable

    """
    def __init__(self, parent):
        KListView.__init__(self, parent)
        self.configList()

    def configList(self):
        self.setSorting(-1)
        
    def __iter__(self):
        """ iterate over QListView items

        """
        it = QListViewItemIterator(self)
        item = it.current()
        while item:
            yield item
            it += 1            
            item = it.current()


class ShortListView(ListView):
    """ ShortListView -> KListView with height set equal to a number of items

    """
    def __init__(self, parent, text, lines):
        ListView.__init__(self, parent)
        self.addColumn(text)
        self.setResizeMode(self.LastColumn)
        self.setHeightLines(lines)
        
    def setHeightLines(self, num):
        height = self.fontMetrics().lineSpacing() * num
        height += self.header().height()
        self.setMaximumHeight(height)


class BasicListView(ListView):
    """ BasicListView -> KListView configured for a single column, no
        sorting, hidden header and no horizontal scrollbar.

    """
    def configList(self):
        self.addColumn('')
        self.setSorting(-1)
        self.header().hide()
        self.setHScrollBarMode(self.AlwaysOff)


class WidgetStack(QWidgetStack):
    """ WidgetStack -> QWidgetStack with QVBoxLayout

    """
    def __init__(self, parent):
        QWidgetStack.__init__(self, parent)
        layout = QVBoxLayout(self)
        layout.setAutoAdd(True)


class GroupBox(QGroupBox):
    """ GroupBox -> QGroupBox with nicer parameters

    """
    def __init__(self, strips, orient, text, parent, margin=0, spacing=0):
        QGroupBox.__init__(self, strips, orient, text, parent)
        self.setFlat(True)
        self.setInsideMargin(margin)
        self.setInsideSpacing(spacing)    

    def addSpaces(self, count, size=0):
        """ adds a number of spaces to the groupbox

        @param count number of spaces to add
        @param size size of each space
        @return None
        """
        for i in range(count):
            self.addSpace(size)


def guiButton(parent, name, layout=None, enable=True):
    """ builds a standard push button by name

    @param parent parent of button
    @param name name of standard item; refer to KStdGuiItem static members
    @keyparam layout=None if given, new button is added to this layout
    @keyparam enable=True initial enabled state of button
    @return KPushButton instance
    """
    gui = getattr(KStdGuiItem, name)()
    button = KPushButton(gui.iconSet(), gui.text(), parent)
    if layout:
        layout.addWidget(button)
    button.setEnabled(enable)
    return button
